###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################
"""Main factory of Bioconvert"""
import copy
import time
import abc
import select
import sys
import inspect
import shutil
import subprocess
import itertools

from subprocess import Popen, PIPE
from io import StringIO
from collections import deque

from easydev import TempFile
from easydev.multicore import cpu_count

import colorlog

import bioconvert

from bioconvert.core.benchmark import Benchmark
from bioconvert.core import extensions

from bioconvert.core.utils import generate_outfile_name
from bioconvert import logger
from . import BioconvertError


_log = colorlog.getLogger(__name__)


class ConvMeta(abc.ABCMeta):
    """This metaclass checks that the converter classes have

       * an attribute input_ext
       * an attribute output_ext

    This is a meta class used by :class:`ConvBase` class. For developers
    only.
    """

    @classmethod
    def split_converter_to_format(cls, converter_name: str):
        converter_name = converter_name.replace("_to_", "2")
        if "2" not in converter_name:
            raise TypeError(
                "converter's name '{}' name must follow convention input2output".format(
                    converter_name
                )
            )
        # for BZ2 2 GZ
        if "22" in converter_name:
            input_fmt, output_fmt = converter_name.upper().split("22", 1)
            input_fmt += "2"
            input_fmt = tuple([input_fmt])
            output_fmt = tuple([output_fmt])
        else:
            input_fmt, output_fmt = converter_name.upper().split("2", 1)
            input_fmt = input_fmt.upper().split("_")
            input_fmt = tuple(input_fmt)
            output_fmt = output_fmt.upper().split("_")
            output_fmt = tuple(output_fmt)

        return input_fmt, output_fmt

    @classmethod
    def lower_tuple(cls, format_tuple):
        format_tuple = [format.lower() for format in format_tuple]
        return format_tuple

    def __init__(cls, name, bases, classdict):

        # do not check extension since modules does not require to specify
        # extension anymore

        def is_conversion_method(item):
            """Return True if method name starts with _method_

            This method is used to keep methods that starts with _method_.
            It uses inspect.getmembers func to list
            all conversion methods implemented in a convertor class.

            :param item: the object to inspect
            :return: True if method's name starts with '__method_', False otherwise.
            :rtype: boolean
            """

            return inspect.isfunction(item) and item.__name__.startswith("_method_")

        if name != "ConvBase":
            input_fmt, output_fmt = cls.split_converter_to_format(name)
            setattr(cls, "input_fmt", input_fmt)
            setattr(cls, "output_fmt", output_fmt)

            if not cls.input_ext:
                # We add all the extensions for each converter into a list.
                input_ext = []
                cls.input_ext = cls.lower_tuple(cls.input_fmt)
                for format in cls.input_ext:
                    input_ext.append(tuple(extensions.extensions[format]))
                # then we turn the list into tuple as output_ext attribute
                setattr(cls, "input_ext", tuple(input_ext))
            # if the developer did not specify an output_ext attribute
            if not cls.output_ext:
                # We add all the extensions for each converter into a list.
                output_ext = []
                cls.output_ext = cls.lower_tuple(cls.output_fmt)
                for format in cls.output_ext:
                    output_ext.append(tuple(extensions.extensions[format]))
                # then we turn the list into tuple as output_ext attribute
                setattr(cls, "output_ext", tuple(output_ext))
                # if the key is not in the dictionary return an error message
            available_conv_meth = []
            for name in inspect.getmembers(cls, is_conversion_method):
                # do not use strip() but split()
                conv_meth = name[0].split("_method_")[1]
                is_disabled = getattr(name[1], "is_disabled", None)
                if is_disabled is None:
                    _log.debug(
                        "converter '{}': method {} is not decorated, we expect it to work all time".format(
                            cls.__name__, conv_meth,
                        )
                    )
                    is_disabled = False
                if not is_disabled:
                    available_conv_meth.append(conv_meth)
                else:
                    _log.warning(
                        "converter '{}': method {} is not available".format(
                            cls.__name__, conv_meth,
                        )
                    )
            setattr(cls, "available_methods", available_conv_meth)
            _log.debug(
                "class = {}  available_methods = {}".format(
                    cls.__name__, available_conv_meth
                )
            )


class ConvArg(object):
    """This class can be used to add specific extra arguments to any converter

    For instance, imagine a conversion named **A2B** that requires the
    user to provide a reference. Then, you may want to provide the
    `--reference` extra argument. This is possible by adding a class
    method named get_additional_arguments that will yield instance of
    this class for each extra argument.

    ::

        @classmethod
        def get_additional_arguments(cls):
            yield ConvArg(
                names="--reference",
                default=None,
                help="the referenc"
            )

    Then, when calling bioconvert as follows,::

        bioconvert A2B --help

    the new argument will be shown in the list of arguments.


    """

    black_listed_argument_for_argparse = [
        "output_argument",
    ]

    def __init__(self, names, help, **kwargs):
        if isinstance(names, list):
            self.args_for_sub_parser = names
        else:
            self.args_for_sub_parser = [
                names,
            ]
        self.kwargs_for_sub_parser = {"help": help}
        self.kwargs_for_sub_parser.update(kwargs)

    def add_to_sub_parser(self, sub_parser):
        kwargs = copy.deepcopy(self.kwargs_for_sub_parser)
        for a in self.black_listed_argument_for_argparse:
            kwargs.pop(a, None)
        sub_parser.add_argument(*self.args_for_sub_parser, **kwargs)

    @classmethod
    def file(cls, path):
        return path


class ConvBase(metaclass=ConvMeta):
    """Base class for all converters.

    To build a new converter, create a new class which inherits from
    :class:`ConvBase` and implement method that performs the conversion.
    The name of the converter method must start with ``_method_``.

    For instance: ::

        class FASTQ2FASTA(ConvBase):

            def _method_python(self, *args, **kwargs):
                # include your code here. You can use the infile and outfile
                # attributes.
                self.infile
                self.outfile

    """

    # specify the extensions of the input file, can be a sequence (must be
    # overridden in subclasses)
    input_ext = None

    # specify the extensions of the output file, can be a sequence (must be
    # overridden in subclasses)
    output_ext = None

    # list available methods
    available_methods = []

    # default method should be provided
    _default_method = None
    _is_compressor = False
    # Can be overriden and if True, new argument --thread is added automatically
    _threading = False
    _extra_arguments = ""

    # Used for the benchmarking only
    others = {}

    # threads to be used by default if argument is required in a method
    # this will be overriden if _threading set to True and therefore --threads
    # set by the user. It is feed back into Bioconvert class
    threads = cpu_count()

    def __init__(self, infile, outfile):
        """.. rubric:: constructor

        :param str infile: the path of the input file.
        :param str outfile: the path of The output file
        """

        self.infile = infile
        self.outfile = outfile

        # execute mode can be shell or subprocess.
        self._execute_mode = "shell"

        # The logger to be set to INFO, DEBUG, WARNING, ERROR, CRITICAL
        self.logger = logger

    def __call__(self, *args, method_name=None, **kwargs):
        """

        :param str method_name: the method to be found in :attr:`available_methods`
        :param str method: the method to be found in :attr:`available_methods`
        :param *args: positional arguments
        :param *kwargs: keyword arguments

        """
        # If method provided, use it
        if "method" in kwargs:
            method_name = kwargs["method"]
            del kwargs["method"]

        # If not, but there is one argument, presumably this is
        # the method
        method_name = method_name or self.default

        # If not, we need to check the name
        # execute() method for the benchmark
        if method_name not in self.available_methods:
            msg = "Methods available are {}".format(self.available_methods)
            _log.error(msg)
            raise ValueError(msg)

        _log.info("{}> Executing {} method ".format(self.name, method_name))
        # reference to the method requested
        method_reference = getattr(self, "_method_{}".format(method_name))

        # call the method itself

        t1 = time.time()

        method_reference(*args, **kwargs)
        t2 = time.time()
        _log.info("Took {} seconds ".format(t2 - t1))

    @property
    def name(self):
        """
        The name of the class
        """
        return type(self).__name__

    def shell(self, cmd):
        from bioconvert.core.shell import shell

        _log.info("CMD: {}".format(cmd))
        shell(cmd)

    def execute(self, cmd, ignore_errors=False, verbose=False, shell=False):

        if ">" in cmd:
            lhs, rhs = cmd.split(">", 1)
            cmd = lhs + self._extra_arguments + ">" + rhs
        else:
            cmd = cmd + self._extra_arguments

        if shell is True or self._execute_mode == "shell":
            self.shell(cmd)
            return
        self._execute(cmd, ignore_errors, verbose)

    def _execute(self, cmd, ignore_errors=False, verbose=False):
        """
        Execute a command in a sub-shell

        :param str cmd: the command to execute
        :param ignore_errors: If True the result is returned whatever the
                              return value of the sub-shell.
                              Otherwise a Runtime error is raised when the sub-shell
                              return a non zero value
        :param verbose: If true displays errors on standard error
        :return: the result of the command
        :rtype: a :class:`StringIO` instance
        """
        try:
            process_ = Popen(cmd, shell=True, stdout=PIPE, stderr=PIPE, stdin=None)
        except Exception as err:
            msg = "Failed to execute Command: '{}'. error: '{}'".format(cmd, err)
            raise RuntimeError(msg)

        inputs = [process_.stdout, process_.stderr]
        output = StringIO()
        errors = StringIO()
        while process_.poll() is None:
            # select has 3 parameters, 3 lists, the sockets, the fileobject to watch
            # in reading, writing, the errors
            # in addition a timeout option (the call is blocking while a fileObject
            # is not ready to be processed)
            # by return we get 3 lists with the fileObject to be processed
            # in reading, writing, errors.
            readable, writable, exceptional = select.select(inputs, [], [], 1)

            while readable and inputs:
                for flow in readable:
                    data = flow.read()
                    if not data:
                        # the flow ready in reading which has no data
                        # is a closed flow
                        # thus we must stop to watch it
                        inputs.remove(flow)
                    if flow is process_.stdout:
                        output.write(data.decode("utf-8"))
                    elif flow is process_.stderr:
                        errors.write(data.decode("utf-8"))
                        print(process_.stderr)
                readable, writable, exceptional = select.select(inputs, [], [], 1)

        errors = errors.getvalue().strip()
        if verbose:
            if errors:
                print(errors, file=sys.stderr)

        if process_.returncode != 0:
            if not ignore_errors:
                raise RuntimeError(errors)
        else:
            return output

    def boxplot_benchmark(
        self,
        N=5,
        rerun=True,
        to_exclude=[],
        to_include=[],
        rot_xticks=90,
        boxplot_args={},
    ):
        """Simple wrapper to call :class:`Benchmark` and plot the results

        see :class:`~bioconvert.core.benchmark.Benchmark` for details.

        """
        if to_include == "all":
            to_include = []

        self._benchmark = Benchmark(
            self, N=N, to_exclude=to_exclude, to_include=to_include
        )
        data = self._benchmark.plot(
            rerun=rerun, rot_xticks=rot_xticks, boxplot_args=boxplot_args
        )
        return data

    def _get_default_method(self):
        if self._default_method is None:
            return self.available_methods[0]
        elif self._default_method not in self.available_methods:
            return self.available_methods[0]
        else:
            return self._default_method

    default = property(_get_default_method)

    def install_tool(self, executable):
        """Install the given tool, using the script:
        bioconvert/install_script/install_executable.sh
        if the executable is not already present

        :param executable: executable to install
        :return: nothing

        """
        # imported but not unused (when we don't have bioconvert_path)
        # import bioconvert
        from bioconvert import bioconvert_data

        if shutil.which(executable) is None:
            logger.info("Installing tool : " + executable)
            # Assigned but never used, says flake8
            # bioconvert_path = bioconvert.__path__[0]
            script = bioconvert_data("install_" + executable + ".sh", where="../misc")
            subprocess.call(["sh", script])

    @classmethod
    def add_argument_to_parser(cls, sub_parser):
        sub_parser.description = cls.get_description()
        for arg in itertools.chain(
            cls.get_IO_arguments(),
            cls.get_common_arguments_for_converter(),
            cls.get_additional_arguments(),
        ):
            arg.add_to_sub_parser(sub_parser)

    @classmethod
    def get_description(cls):
        msg = "Convert file from '{}' to '{}' format. "
        msg += "See bioconvert.readthedocs.io for details"
        msg = msg.format(*ConvMeta.split_converter_to_format(cls.__name__))
        return msg

    @classmethod
    def get_additional_arguments(cls):
        return []

    # common arguments for the sub command case
    # when using bioconvert <conversion>
    @staticmethod
    def get_IO_arguments():
        yield ConvArg(
            names="input_file",
            nargs="?",
            default=None,
            type=ConvArg.file,
            help="The path to the file to convert.",
        )
        yield ConvArg(
            names="output_file",
            nargs="?",
            default=None,
            type=ConvArg.file,
            output_argument=True,
            help="The path where the result will be stored.",
        )

    @staticmethod
    def get_common_arguments():
        yield ConvArg(
            names=["-f", "--force",],
            action="store_true",
            help="if outfile exists, it is overwritten with this option",
        )
        yield ConvArg(
            names=["-v", "--verbosity",],
            default=bioconvert.logger.level,
            help="Set the outpout verbosity.",
            choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        )
        yield ConvArg(
            names=["--raise-exception",],
            action="store_true",
            help="Let exception ending the execution be raised and displayed",
        )
        yield ConvArg(
            names=["-X", "--batch",],
            default=False,
            action="store_true",
            help="Allow conversion of a set of files using wildcards. You "
            "must use quotes to escape the wildcards. For instance: "
            "--batch 'test*fastq' ",
        )
        yield ConvArg(
            names=["-b", "--benchmark",],
            default=False,
            action="store_true",
            help="Running all available methods",
        )
        yield ConvArg(
            names=["-N", "--benchmark-N",],
            default=5,
            type=int,
            help="Number of trials for each methods",
        )
        yield ConvArg(
            names=["-T", "--benchmark-tag",],
            default="bioconvert",
            help="Save results (json and image) named after this tag",
        )
        yield ConvArg(
            names=["-I", "--benchmark-save-image",],
            action="store_true",
            help="Save results in this image",
        )
        yield ConvArg(
            names=["-M", "--benchmark-methods",],
            default="all",
            nargs="+",
            type=str,
            help="Methods to include. Provide list as space-separated method names. Use -s  to get the full list.",
        )
        yield ConvArg(
            names=["-a", "--allow-indirect-conversion",],
            default=False,
            action="store_true",
            help="Allow to chain converter when direct conversion is absent",
        )
        yield ConvArg(
            names=["-e", "--extra-arguments",],
            default="",
            help="Any arguments accepted by the method's tool",
        )

    @classmethod
    def get_common_arguments_for_converter(cls):
        for a in ConvBase.get_common_arguments():
            yield a
        try:
            # Some converters do not have any method and work
            # in __call__, so preventing to crash by searching for them
            yield ConvArg(
                names=["-m", "--method",],
                nargs="?",
                default=cls._get_default_method(cls),
                help="The method to use to do the conversion.",
                choices=cls.available_methods,
            )
        except Exception as e:
            _log.warning(
                "converter '{}' does not seems to have methods: {}".format(
                    cls.__name__, e
                )
            )
            pass
        yield ConvArg(
            names=["-s", "--show-methods",],
            default=False,
            action="store_true",
            help="A converter may have several methods",
        )

        if cls._threading:
            yield ConvArg(
                names=["-t", "--threads"],
                # nargs=1,
                type=int,
                default=cls.threads,
                help="threads to be used",
            )


# Implementing a class creator
# The created class will have the correct name, will inherit from ConvBase
# It will have a conversion method chaining conversions through tempfiles
def make_chain(converter_map):
    """
    Create a class performing step-by-step conversions following a path.
    *converter_map* is a list of pairs ((in_fmt, out_fmt), converter).
    It describes the conversion path.
    """
    in_fmt = converter_map[0][0][0]
    out_fmt = converter_map[-1][0][1]
    chain_name = "{}2{}".format("_".join(in_fmt), "_".join(out_fmt))
    chain_attributes = {}

    def chain_init(self, infile, outfile):
        super().__init__(infile, outfile)
        self._default_method = "chain"

    def _method_chain(self, *args, **kwargs):
        """This method successively uses the default conversion method of each
        converter in the conversion path."""

        def conv_step(converter, infile, outfile):
            """Performs one conversion step."""
            converter(infile, outfile)(*args, **kwargs)

        # Contains the last temporary output file, if any
        pipe_files = deque()
        for (step_num, ((_, out_fmt), converter)) in enumerate(
            self.converter_map, start=1
        ):
            if step_num == 1:
                # May not be necessary:
                step_infile = None
                step_input = self.infile
                del_infile = False
            else:
                step_infile = pipe_files.popleft()
                step_input = step_infile.name
                del_infile = True

            if step_num == self.nb_steps:
                # May not be necessary:
                step_outfile = None
                step_output = self.outfile
            else:

                # FIXME: for mutiple IO converters
                if len(out_fmt) == 1:
                    step_outfile = TempFile(suffix=out_fmt[0].lower())
                    step_output = step_outfile.name
                    pipe_files.append(step_outfile)

            conv_step(converter, step_input, step_output)
            if del_infile:
                step_infile.delete()

    chain_attributes["converter_map"] = converter_map
    chain_attributes["nb_steps"] = len(converter_map)
    chain_attributes["__init__"] = chain_init
    chain_attributes["_method_chain"] = _method_chain
    chain = type(chain_name, (ConvBase,), chain_attributes)
    # https://stackoverflow.com/a/43779009/1878788
    # Allows calling super in chain.__init__
    __class__ = chain
    return chain



import os
import re
import sys

## used for nt reverse complements
_nt_comp_table = bytes.maketrans(b'ACBDGHKMNSRUTWVYacbdghkmnsrutwvy', \
                                 b'TGVHCDMKNSYAAWBRtgvhcdmknsyaawbr')

# for translations: http://www.ncbi.nlm.nih.gov/Taxonomy/Utils/wprintgc.cgi
_translation_table = {
    1: { 
        'GCA':'A', 'GCC':'A', 'GCG':'A', 'GCT':'A',
        'CGA':'R', 'CGC':'R', 'CGG':'R', 'CGT':'R', 'AGA':'R', 'AGG':'R',
        'AAC':'N', 'AAT':'N',
        'GAC':'D', 'GAT':'D', 
        'TGC':'C', 'TGT':'C',
        'GAA':'E', 'GAG':'E',
        'CAA':'Q', 'CAG':'Q',
        'GGA':'G', 'GGC':'G', 'GGG':'G', 'GGT':'G',
        'CAC':'H', 'CAT':'H',
        'ATA':'I', 'ATC':'I', 'ATT':'I',
        'TTA':'L', 'TTG':'L', 'CTA':'L', 'CTC':'L', 'CTG':'L','CTT':'L',
        'AAA':'K', 'AAG':'K',
        'ATG':'M',
        'TTC':'F', 'TTT':'F',
        'CCA':'P', 'CCC':'P', 'CCG':'P', 'CCT':'P',
        'TCA':'S', 'TCC':'S', 'TCG':'S', 'TCT':'S', 'AGC':'S', 'AGT':'S',
        'ACA':'T', 'ACC':'T', 'ACG':'T', 'ACT':'T',
        'TGG':'W',
        'TAC':'Y', 'TAT':'Y',
        'GTA':'V', 'GTC':'V', 'GTG':'V', 'GTT':'V',
        'TAA':'*', 'TAG':'*', 'TGA':'*'
        }
}

def translate( seq, translation_table=None ):
    """
    Does a direct translation of the passed DNA/RNA sequence in phase 0.
    You can pass a numeric translation table, else 1 is assumed.

    An 'X' is used when a codon is unknown (contains an N)

    http://www.ncbi.nlm.nih.gov/Taxonomy/Utils/wprintgc.cgi
    """
    if translation_table is None:
        translation_table = 1

    # make sure we've defined this translation table
    if translation_table not in _nt_comp_table:
        raise Exception("ERROR: translation requested using table {0}, which isn't yet supported.".format(translation_table))

    trans_table = _translation_table[translation_table]
    
    # In case an RNA string was passed
    seq = seq.translate(seq.maketrans('Uutagc', 'TTTAGC'))

    polypeptide_seq = ''
    x = 0

    while True:
        try:
            polypeptide_seq += trans_table[seq[x:x+3]]
        except (IndexError):
            break
        except (KeyError):
            if len(seq[x:x+3]) == 3:
                #raise Exception("ERROR: Encountered unknown codon during translation: {0}".format(seq[x:x+3]))
                print("WARN: Encountered unknown codon during translation: {0}".format(seq[x:x+3]))
                polypeptide_seq += 'X'
            else:
                break
        x += 3
        
    
    return polypeptide_seq


def read_list_file( file ):
    """Parse an list file and return an array of the paths"""
    files = []

    if ( not os.path.isfile(file) ):
        raise Exception("Couldn't find file: " + file)

    ## only do non-blank lines
    with open(file) as f_in:
        lines = filter(None, (line.rstrip() for line in f_in))

        for line in lines:
            files.append(line)

    return files


def reverse_complement( seq ):
    """
    Biological reverse complementation.  Case in sequences are retained, and 
    IUPAC codes are supported.  Code modified from:

    http://shootout.alioth.debian.org/u32/program.php?test=revcomp&lang=python3&id=4
    """
    return seq.translate(_nt_comp_table)[::-1]
    

def humancoords_to_0interbase( start, stop ):
    """
    The typical human-readable coordinate system, such as found in GBK flat files,
    has a start and stop coordinate only.  They are 1-based, on-base coordinates
    and features on a reverse strand are indicated by having start > stop.  This
    transforms them into the GMOD standard 0-based inter-base coordinates.

    Returns a list of fmin, fmax and strand values.
    """
    fmin = start
    fmax = stop
    strand = 1
    
    if ( stop < start ):
        fmin = stop
        fmax = start
        strand = -1

    fmin -= 1

    return (fmin, fmax, strand)

def interbase0_to_humancoords( fmin, fmax, strand ):
    """
    Takes GMOD-standard 0-based inter-base coordinates and transforms them
    into the typical human-readable coordinates used in places like GBK flat
    files.  Features on the reverse strand have a larger start than stop.

    The strand values can be:

    + or 1 for forward features
    - or -1 for reverse features

    Returns a list: [start, stop] where start < stop if on forward strand.
    """

    if strand == '+' or strand == 1:
        return (fmin + 1, fmax)
    elif strand == '-' or strand == -1:
        return (fmax, fmin + 1)
    else:
        raise Exception("Invalid strand specified ({0}).  Expected +, -, 1 or -1".format(strand))


def fasta_dict_from_file( file ):
    """
    Reads a file of FASTA entries and returns a dict where each key is a sequence ID.
    The value is another dict with two keys 'h' for header and 's' for sequence.  The
    header is all the other text after the id in the original FASTA header.  The
    sequence has all whitespace removed.  Obviously this should only be used on files
    where memory to load them isn't an issue.
    """
    seqs = dict()
    current_seq = ''
    current_id = None
    current_header = None
    
    for line in open(file):
        line = line.rstrip()
        m = re.search('>(\S+)\s*(.*)', line)
        if m:
            ## new residue line matched, purge the existing one, if not the first
            if current_id is not None:
                ## warn if it has already been found
                if current_id in seqs:
                    sys.stderr.write("WARN: Duplicate ID ({0}) found.  Only last one kept.\n".format(current_id))

                ## remove all whitespace and save
                current_seq = ''.join(current_seq.split())
                seqs[current_id] = {'h':current_header, 's':current_seq}
                    
            current_seq = ''
            current_id = m.group(1)
            current_header = m.group(2)
        else:
            ## python 2.6+ makes string concatenation amortized O(n)
            ##  http://stackoverflow.com/a/4435752/1368079
            current_seq += str(line)

    ## don't forget the last one
    current_seq = ''.join(current_seq.split())
    seqs[current_id] = {'h':current_header, 's':current_seq}

    return seqs


def fasta_sizes_from_file(file):
    """
    Reads a file of FASTA entries and returns a dict where each key is a sequence ID and
    the value is just the size of each sequence.  This is as almost as computationally
    intensive as fasta_dict_from_file() but takes less memory and is appropriate when you
    only care about the residue lengths.
    """
    seqs = dict()
    current_id = None
    
    for line in open(file):
        line = line.rstrip()
        m = re.search('>(\S+)\s*(.*)', line)
        if m:
            ## new residue line matched, set the new seq ID
            current_id = m.group(1)
            seqs[current_id] = 0
        else:
            seqs[current_id] += len(line)

    return seqs


def add_assembly_fasta(mols, fasta_file):
    """
    Takes a collection of molecules (Assembly) objects and adds sequence residues based on
    their IDs being found in the passed FASTA file.
    """
    fasta_seqs = fasta_dict_from_file(fasta_file)

    for mol_id in mols:
        # check if the FASTA file provides sequence for this
        if mol_id in fasta_seqs:
            mol = mols[mol_id]
            mol.residues = fasta_seqs[mol_id]['s']
            mol.length   = len(mol.residues)
            
def wrapped_fasta(string, every=60):
    """
    Pass a string of residues (nucleotide or polypeptides) that has NO whitespace and
    this will return another string with new line characters inserted every N residues.

    N is specified with the 'every' parameter with a default of 60.  For example:

    new_fasta = wrapped_fasta(some_string, every=60)
    
    This runs orders of magnitude faster than using the textwrap module.
    """
    return '\n'.join(string[i:i+every] for i in range(0, len(string), every))



###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################
"""Convert :term:`JSON` to :term:`YAML` format"""
import yaml
import json
from bioconvert import ConvBase
from bioconvert.core.decorators import requires, requires_nothing, compressor


__all__ = ["JSON2YAML"]


class JSON2YAML(ConvBase):
    """Convert :term:`JSON` file into :term:`YAML` file

    Conversion is based on yaml and json standard Python modules
    Indentation is set to 4 by default and affects the sections (not the list).
    For example::

        fruits_list:
        - apple
        - orange
        section1:
            do: true
            misc: 1

    """

    #: Default value
    _default_method = "yaml"

    def __init__(self, infile, outfile, *args, **kargs):
        """.. rubric:: constructor

        :param str infile: input JSON file
        :param str outfile: input YAML file. 
        """
        super(JSON2YAML, self).__init__(infile, outfile, *args, **kargs)

    @requires_nothing
    @compressor
    def _method_yaml(self, *args, **kwargs):
        """Internal method"""
        with open(self.infile, "r") as infile:
            data = json.load(infile)
        with open(self.outfile, "w") as outfile:
            yamldata = yaml.dump(
                data, Dumper=yaml.dumper.Dumper, default_flow_style="", indent=4
            )
            outfile.write(yamldata)



###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################
"""Provides a general tool to perform pre/post compression"""
from distutils.spawn import find_executable
from functools import wraps
from os.path import splitext

import colorlog
import pkg_resources
from easydev import TempFile

_log = colorlog.getLogger(__name__)


def in_gz(func):
    """Marks a function as accepting gzipped input."""
    func.in_gz = True
    return func


def make_in_gz_tester(converter):
    """Generates a function testing whether a conversion method of *converter*
    has the *in_gz* tag."""

    def is_in_gz(method):
        """Accesses the function corresponding to *method* and tells whether it
        has the *in_gz* tag."""
        return hasattr(getattr(converter, "_method_{}".format(method)), "in_gz")

    return is_in_gz


def compressor(func):
    """Decompress/compress input file without pipes

    Does not use pipe: we decompress and compress back the input file.
    The advantage is that it should work for any files (even very large).

    This decorator should be used by method that uses pure python code
    """

    # https://stackoverflow.com/a/309000/1878788
    @wraps(func)
    def wrapped(inst, *args, **kwargs):
        infile_name = inst.infile
        if type(inst.outfile) is not list:

            output_compressed = None
            if inst.outfile.endswith(".gz"):
                (inst.outfile, output_compressed) = splitext(inst.outfile)
            elif inst.outfile.endswith(".bz2"):
                (inst.outfile, output_compressed) = splitext(inst.outfile)
            elif inst.outfile.endswith(".dsrc"):  # !!! only for fastq files
                (inst.outfile, output_compressed) = splitext(inst.outfile)
            # Now inst has the uncompressed output file name

        if infile_name.endswith(".gz"):
            # decompress input
            # TODO: https://stackoverflow.com/a/29371584/1878788
            _log.info("Generating uncompressed version of {} ".format(infile_name))
            (ungz_name, _) = splitext(infile_name)
            (_, base_suffix) = splitext(ungz_name)
            with TempFile(suffix=base_suffix) as ungz_infile:
                inst.infile = ungz_infile.name
                inst.shell(
                    "unpigz -c -p {} {} > {}".format(
                        inst.threads, infile_name, inst.infile
                    )
                )
                # computation
                results = func(inst, *args, **kwargs)
            inst.infile = infile_name
        else:
            results = func(inst, *args, **kwargs)

        # Compress output and restore inst output file name
        if output_compressed == ".gz":
            # TODO: this uses -f ; should be a
            _log.info("Compressing output into .gz")
            inst.shell("pigz -f -p {} {}".format(inst.threads, inst.outfile))
            inst.outfile = inst.outfile + ".gz"
        elif output_compressed == ".bz2":
            _log.info("Compressing output into .bz2")
            inst.shell("pbzip2 -f -p{} {}".format(inst.threads, inst.outfile))
            inst.outfile = inst.outfile + ".bz2"
        elif output_compressed == ".dsrc":  # !!! only for FastQ files
            _log.info("Compressing output into .dsrc")
            inst.shell(
                "dsrc c -t{} {} {}.dsrc".format(
                    inst.threads, inst.outfile, inst.outfile
                )
            )
            inst.outfile = inst.outfile + ".dsrc"
        return results

    return in_gz(wrapped)


def out_compressor(func):
    """Compress output file without pipes

    This decorator should be used by method that uses pure python code
    """

    # https://stackoverflow.com/a/309000/1878788
    @wraps(func)
    def wrapped(inst, *args, **kwargs):
        output_compressed = None
        if inst.outfile.endswith(".gz"):
            (inst.outfile, output_compressed) = splitext(inst.outfile)
        elif inst.outfile.endswith(".bz2"):
            (inst.outfile, output_compressed) = splitext(inst.outfile)
        elif inst.outfile.endswith(".dsrc"):  # !!! only for fastq files
            (inst.outfile, output_compressed) = splitext(inst.outfile)
        # Now inst has the uncompressed output file name

        # computation
        results = func(inst, *args, **kwargs)

        # Compress output and restore inst output file name
        if output_compressed == ".gz":
            # TODO: this uses -f ; should be a
            _log.info("Compressing output into .gz")
            inst.shell("pigz -f -p {} {}".format(inst.threads, inst.outfile))
            inst.outfile = inst.outfile + ".gz"
        elif output_compressed == ".bz2":
            _log.info("Compressing output into .bz2")
            inst.shell("pbzip2 -f -p{} {}".format(inst.threads, inst.outfile))
            inst.outfile = inst.outfile + ".bz2"
        elif output_compressed == ".dsrc":  # !!! only for FastQ files
            _log.info("Compressing output into .dsrc")
            inst.shell(
                "dsrc c -t{} {} {}.dsrc".format(
                    inst.threads, inst.outfile, inst.outfile
                )
            )
            inst.outfile = inst.outfile + ".dsrc"
        return results

    return wrapped


def requires_nothing(func):
    """Marks a function as not needing dependencies."""
    func.is_disabled = False
    return func


def requires(
    external_binary=None,
    python_library=None,
    external_binaries=None,
    python_libraries=None,
):
    """

    :param external_binary: a system binary required for the method
    :param python_library:  a python library required for the method
    :param external_binaries: an array of system binaries required for the method
    :param python_libraries: an array of python libraries required for the method
    :return:
    """
    external_binaries = external_binaries or []
    python_libraries = python_libraries or []
    if external_binary:
        external_binaries.append(external_binary)
    if python_library:
        python_libraries.append(python_library)

    __missing_binaries = getattr(requires, "__missing_binaries", {})
    requires.__missing_binaries = __missing_binaries
    __missing_libraries = getattr(requires, "__missing_libraries", {})
    requires.__missing_libraries = __missing_libraries
    __pip_libraries = getattr(requires, "__pip_libraries", None)
    if __pip_libraries is None:
        __pip_libraries = [p.project_name for p in pkg_resources.working_set]
        requires.__pip_libraries = __pip_libraries

    def real_decorator(function):
        @wraps(function)
        def wrapped(inst, *args, **kwargs):
            return function(inst, *args, **kwargs)

        try:
            for bin in external_binaries:
                try:
                    if __missing_binaries[bin]:
                        raise Exception("{} has already be seen as missing".format(bin))
                except KeyError:
                    __missing_binaries[bin] = True
                    # shell("which %s" % bin)
                    if find_executable(bin) is None:
                        raise Exception("{} was not found in path".format(bin))
                    __missing_binaries[bin] = False
            for lib in python_libraries:
                try:
                    if __missing_libraries[lib]:
                        raise Exception("{} has already be seen as missing".format(lib))
                except KeyError:
                    missing = lib not in __pip_libraries
                    __missing_libraries[lib] = missing
                    if missing:
                        raise Exception("{} was not found by pip".format(lib))
            wrapped.is_disabled = False
        except Exception as e:
            _log.debug(e)
            wrapped.is_disabled = True
        return wrapped

    return real_decorator


def get_known_dependencies_with_availability(as_dict=False):
    if as_dict:
        external_binaries = {}
        python_libraries = {}
        for binary, missing in getattr(requires, "__missing_binaries", {}).items():
            external_binaries[binary] = dict(available=not missing,)
        for library, missing in getattr(requires, "__missing_libraries", {}).items():
            python_libraries[library] = dict(available=not missing,)
        return dict(
            external_binaries=external_binaries, python_libraries=python_libraries,
        )
    ret = []
    for binary, status in sorted(getattr(requires, "__missing_binaries", {}).items()):
        ret.append((binary, not status, "binary",))
    for library, status in sorted(getattr(requires, "__missing_libraries", {}).items()):
        ret.append((library, not status, "library",))
    return ret



# -*- coding: utf-8 -*-

###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################

"""
Possible Conversion
====================

Plot directed graph of possible conversion


"""
#################################################
#
from bioconvert.core.graph import create_graph


#####################################################
# If you use pygraphviz, you can have a good quality
# image using:
import matplotlib as mpl

mpl.rcParams["figure.dpi"] = 250

############################################################
# In order to create the following image, you need graphviz
# and pygraphviz.
# If you cannot install those packages, you may use a singularity
# image like in the following example by setting the use_singularity
# parameter to True. This would work under Linux. Not tested on other systems
# yet.
try:
    create_graph("conversion.png", use_singularity=True)
except:
    create_graph("conversion.png", use_singularity=False)


#####################################################
#
from pylab import imshow, imread, xticks, yticks, gca

imshow(imread("conversion.png"), interpolation="nearest")
xticks([])
yticks([])
ax = gca()
ax.axis("off")



###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################
"""Convert :term:`BAM` file to :term:`TSV` format"""
import os
from bioconvert import ConvBase
from bioconvert.core.decorators import requires

import colorlog

logger = colorlog.getLogger(__name__)


class BAM2TSV(ConvBase):
    """Convert sorted :term:`BAM` file into :term:`TSV` stats

    This is not a conversion per se but the extraction of BAM
    statistics saved into a TSV format. The 4 columns of the TSV file
    are::

        Reference sequence name, Sequence length,Mapped reads, Unmapped reads


    Methods are based on samtools [SAMTOOLS]_ and pysam [PYSAM]_.
    """

    #: Default value
    _default_method = "samtools"

    def __init__(self, infile, outfile, *args, **kargs):
        """.. rubric:: constructor

        :param str infile: BAM file
        :param str outfile: TSV file

        Methods are based on samtools [SAMTOOLS]_ and pysam [PYSAM]_.

        """
        super(BAM2TSV, self).__init__(infile, outfile, *args, **kargs)

    @requires("samtools")
    def _method_samtools(self, *args, **kwargs):
        """Here we use the SAMtools tool.

        `SAMtools documentation <http://www.htslib.org/doc/samtools.html>`_"""
        with open(self.outfile, "wt") as out:
            out.write(
                "Reference sequence name\tSequence length\t"
                "Mapped reads\tUnmapped reads{}".format(os.linesep)
            )
        cmd = "samtools index {0} && samtools idxstats {0} >> {1}".format(
            self.infile, self.outfile
        )
        self.execute(cmd)

    @requires(python_library="pysam", external_binary="samtools")
    def _method_pysam(self, *args, **kwargs):
        """We use here the python module Pysam.

        `Pysam documentation <https://pysam.readthedocs.io/en/latest/api.html>`_"""
        import pysam

        # index the bam file
        pysam.index(self.infile)
        # create count table
        with open(self.outfile, "wt") as out:
            out.write(
                "Reference sequence name\tSequence length\t"
                "Mapped reads\tUnmapped reads{}".format(os.linesep)
            )
            for line in pysam.idxstats(self.infile):
                out.write(line)



###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################
"""Convert :term:`SRA` format to :term:`FASTA` format"""

from bioconvert import ConvBase
import subprocess
import os
import tempfile
import shutil

from bioconvert.core.decorators import requires


class SRA2FASTQ(ConvBase):
    """Download FASTQ from SRA archive

    ::

        bioconvert sra2fastq ERR043367

    This may take some times since the files are downloaded from SRA website.

    """

    #: Default value
    _default_method = "sratoolkit"

    # If test: will take only the first 10 reads from the sra file
    def __init__(self, infile, outfile, test=False):
        """.. rubric:: constructorhttps://edwards.flinders.edu.au/fastq-dump/

        library used: sra-toolkit
        """
        super().__init__(infile, outfile)
        self.test = test

    @requires("fastq-dump")
    def _method_sratoolkit(self, *args, **kwargs):
        """Uses Sratoolkit (fastq-dump) to convert a sra file to fastq

        `Fastq-dump documentation <https://edwards.flinders.edu.au/fastq-dump/>`_"""
        inname = os.path.split(os.path.splitext(self.infile)[0])[1]
        outbasename, ext = os.path.splitext(self.outfile)
        compresscmd = ""
        gzext = ""
        if ext == ".gz":
            compresscmd = "--gzip"
            gzext = ".gz"
            outbasename = os.path.splitext(outbasename)[0]

        infile = self.infile
        # If the file does not exist locally, we take the basename
        # it should correspond to a SRA ID
        if os.path.isfile(infile) is False:
            infile = inname

        tmpdir = tempfile.mkdtemp()
        testcmd = ""
        # If in test mode, we retrieve only 10 reads from sra
        if self.test:
            testcmd = "-X 10"
        if self.isPairedSRA(infile):
            cmd = "fastq-dump {} {} --split-files -O {} {}".format(
                testcmd, compresscmd, tmpdir, infile
            )
            self.execute(cmd)
            cmd = "mv {}/{}_1.fastq{} {}_1.fastq{}".format(
                tmpdir, inname, gzext, outbasename, gzext
            )
            self.execute(cmd)
            cmd = "mv {}/{}_2.fastq{} {}_2.fastq{}".format(
                tmpdir, inname, gzext, outbasename, gzext
            )
            self.execute(cmd)
        else:
            cmd = "fastq-dump {} {} -O {} {}".format(
                testcmd, compresscmd, tmpdir, infile
            )
            self.execute(cmd)
            cmd = "mv {}/{}.fastq{} {}".format(tmpdir, inname, gzext, self.outfile)
            self.execute(cmd)
        shutil.rmtree(tmpdir)

    def isPairedSRA(self, filename):
        try:
            contents = subprocess.check_output(
                ["fastq-dump", "-X", "1", "-Z", "--split-spot", filename]
            )
        except subprocess.CalledProcessError:
            raise Exception("Error running fastq-dump on", filename)

        if contents.count(b"\n") == 4:
            return False
        elif contents.count(b"\n") == 8:
            return True
        else:
            raise Exception("Unexpected output from fast-dump on ", filename)



###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################
"""Convert :term:`BAM` file to :term:`BIGWIG` format"""

import colorlog
from bioconvert import ConvBase
from bioconvert.core.base import ConvArg


from easydev import TempFile

from bioconvert.core.decorators import requires

_log = colorlog.getLogger(__name__)

__all__ = ["BAM2BIGWIG"]


class BAM2BIGWIG(ConvBase):
    """Convert :term:`BAM` file to :term:`BIGWIG` file

    Convert BAM into a binary version of :term:`WIG` format.

    Methods are base on bamCoverage [DEEPTOOLS]_ and bedGraphToBigWig from
    wiggletools [WIGGLETOOLS]_. Wiggletools method requires an extra argument
    (--chrom-sizes) therefore default one is bamCoverage for now.

    Moreover, the two methods do not return exactly the same info!

    You can check this by using bioconvert to convert into a human readable file
    such as wiggle. We will use the bamCoverage as our default conversion.

    """

    #: Default value
    _default_method = "bamCoverage"

    def __init__(self, infile, outfile, *args, **kargs):
        """.. rubric:: constructor

        :param str infile: input BAM file
        :param str outfile: output BIGWIG filename

        """
        super(BAM2BIGWIG, self).__init__(infile, outfile, *args, **kargs)

    @requires("bamCoverage")
    def _method_bamCoverage(self, *args, **kwargs):
        """run bamCoverage package.
        
        `bamCoverage documentation <https://deeptools.readthedocs.io/en/develop/content/tools/bamCoverage.html>`_"""
        cmd = "bamCoverage --bam {} --outFileFormat bigwig --outFileName {}".format(
            self.infile, self.outfile
        )
        self.execute(cmd)

    @requires(external_binaries=["bedGraphToBigWig", "bedtools"])
    def _method_ucsc(self, *args, **kwargs):
        """Run ucsc tool bedGraphToBigWig. 

        Requires extra argument (chrom_sizes) required by the bioconvert
        stanalone.
        """
        from bioconvert.bam2bedgraph import BAM2BEDGRAPH
        from bioconvert.bedgraph2bigwig import BEDGRAPH2BIGWIG

        chrom_sizes = kwargs.get("chrom_sizes", None)

        with TempFile(suffix=".bedgraph") as fh:
            convertbam2bed = BAM2BEDGRAPH(self.infile, fh.name)
            convertbam2bed()
            convertbed2bw = BEDGRAPH2BIGWIG(fh.name, self.outfile)
            convertbed2bw(chrom_sizes=chrom_sizes)

    @classmethod
    def get_additional_arguments(cls):
        yield ConvArg(
            names="--chrom-sizes",
            default=None,
            help="a two-column file/URL: <chromosome name> <size in bases>. "
            "Used by the ucsc method only",
        )
